PROGRAM Utter;

(* UTTER 1.0, Copyright (c) 1994, by Jody R. Cairns

   UTTER displays quotes that are stored in a separate ASCII data file, by
   default called UTTER.DAT, although a different file can be used if
   specified on the command-line. All quotes must end with a tilde (~).
   The quote file may be edited using the /E option. There are many more
   options (and more to come) including: display random quote, prompt
   user after quote is displayed, change video output, show number of
   quotes in quote file, etc.

   The source to this program (this file) is free to the public domain with
   the condition that I am notified of any modifications made to this
   program, and the executable of the MODIFIED program is NOT distributed
   without my permission (i.e. Don't pull a fast one like making a few
   changes to the source, compiling it, and saying you wrote it.).  Otherwise,
   the source and executable are free to those who want it.

   It would be very much appreciated if any code used from this program is
   credited to me (bearing in mind that I take absolutely no responsibility
   for the results or misuse of the execution or use of said code).

   Main Algorithm
   --------------
      1) Check if valid DOS version (must be >= 3.0).
      2) Get executable's filename and residing directory.
      3) Process command-line for options.
      4) Check if data file containing quotes exists.
      5) Check if index file for data file exists and is valid.
      6) Read index file to get old data file time-stamp.
      7) Compare old data file's time-stamp with current data file's time-stamp.
      8) Update index file if any discrepancies.
      9) Execute options if specified.
     10) Seek into proper location of data file for quote (determined
         from location read from index file).
     11) Read and display quote.
     12) Update index file with next quote to be displayed location.
     13) Close files.

   See PROCEDURE DisplayHelp for more information about options available
   and the data file format.

   See PROCEDURE UpdateIndexFile for information on the format of the
   index file(s).

   NOTE: this program is not very modular due to the speed factor. I wanted
         to make this as fast as possible without adding much assembly code
         and still have sufficient error-checking.  Too many procedures
         would have slowed it down (I timed it). So, please excuse this
         programming oversight; it was done for speed. However, you will
         notice I could have been LESS modular; code size was a factor, too.

   UTTER.PAS was written and compiled using TURBO PASCAL 6.0 and tested
   on a variety of Personal Computers, from 8086s to 80486s.

   If you find ANY bugs, or have any suggestions or ideas I'd be happy to
   hear from you.  Please don't hesitate to contact me for any information.
   I have more utilites in the works, all DOS-based. Here's my address:

   Jody R. Cairns
   P.O. Box 5991
   St. John's, NF
   Canada
   A1C 5X4

   e-mail: jodyc@cs.mun.ca

                                                                          *)

(* Compiler directives follow: *)

{$M 5120,0,0}    (* Memory requirements: 5 KB for stack, 0 KB for heap *)

{$I-}            (* I/O checking off                 *)
{$R-}            (* Range checking off               *)
{$S-}            (* Stack checking off               *)
{$X-}            (* Extended syntax off              *)
{$V-}            (* Strict VAR-string checking off   *)
{$B-}            (* Short-circuit boolean evaluation *)
{$D-}            (* Debug information off            *)
{$L-}            (* Local debug information off      *)
{$E-}            (* 8087 emulation off               *)
{$N-}            (* 8087/80287 code generation off   *)
{$G-}            (* 80286 code generation off        *)
{$F-}            (* Force far calls off              *)
{$A+}            (* Word align data                  *)
{$O-}            (* Overlays not allowed             *)


USES
  DOS, CRT;


LABEL
  EndLoop1,      (* Used for goto statements *)
  EndLoop2;


CONST
  Title         = 'UTTER 1.0, Copyright (c) 1994, Jody R. Cairns';

  EndOfQuote    = '~';        (* End of quote indicator                  *)
  IndexExt      = '.IDX';     (* extension of index file                 *)
  ESC           = #27;        (* ASCII code for <ESC> key                *)
  BufferSize    = 1024 * 61;  (* Maximum size of buffer                  *)
  MaxWord       = 65535;      (* Largest possible word value             *)
  MaxSeconds    = 60;         (* Maximum # of seconds for /P option      *)


TYPE
  NumStr     = string[11];                    (* Longint type converted to string *)
  IDXFile    = file of longint;               (* Format of index file             *)
  BufferType = ARRAY[1..BufferSize] OF char;  (* Buffer used for reading quotes   *)


CONST
  UpdateFile   : boolean = FALSE;  (* Update index file                    *)
  ShowNumber   : boolean = FALSE;  (* Show number of quotes in quote file  *)
  GetRandom    : boolean = FALSE;  (* Get random quote from quote file     *)
  TitleNotShown: boolean = TRUE;   (* Program name has been displayed      *)
  InvokeEditor : boolean = FALSE;  (* Invoke editor to edit file of quotes *)
  DoPrompts    : boolean = TRUE;   (* Display warning prompts              *)
  KeyPause     : boolean = FALSE;  (* Pause with key prompt after quote    *)
  MousePause   : boolean = FALSE;  (* Pause with mouse prompt after quote  *)
  Pause        : boolean = FALSE;  (* Pause for some seconds after quote   *)
  Editor       : PathStr = 'EDIT'; (* Default editor name                  *)
  CurrRow      : byte    = 0;      (* Current row number of screen         *)
  PauseMilSecs : word    = 2000;   (* # of milliseconds to pause for /P    *)
  Prompt       : ComStr  = #13#10'Press any key to continue...';


VAR
  DataFile  : file;       (* File containing quotes                          *)
  IndexFile : IDXFile;    (* Index file for DataFile                         *)
  Buffer    : BufferType; (* Buffer used for reading in quotes               *)
  Dir       : DirStr;     (* Directory that contains UTTER.EXE               *)
  Name      : NameStr;    (* Name of executing program                       *)
  Ext       : ExtStr;     (* Extension of file that contains quotes          *)
  TmpStr,                 (* Used for option /P[x] -> converted number       *)
  Param     : ComStr;     (* Paramater on command-line                       *)
  EditorPath,             (* Full path name of editor                        *)
  DataPath  : PathStr;    (* Full PATH location of file that contains quotes *)
  LongIntVar,             (* Used for the /P option in val procedure         *)
  DataTime,               (* Current time stamp of file containing quotes    *)
  OldDataTime,            (* Old time stamp of file containing quotes        *)
  StartPos  : longint;    (* Index of quotes in Datafile                     *)
  RecsRead,               (* # of records (chars) read into Buffer           *)
  I, J,                   (* Indices and other stuff                         *)
  Code      : word;       (* Temporary storage                               *)
  Ch        : char;       (* Character pressed during prompt                 *)
  NumRows   : byte;       (* Number of rows on screen                        *)
  Reg       : registers;  (* CPU registers                                   *)


(*****************************************************************************)
(*****************************************************************************)


FUNCTION KeyPressed: boolean; assembler;
(* Returns TRUE if a key is pressed. *)
  ASM
    mov ah, 11h     (* Report whether character is ready *)
    int 16h         (* Call Keyboard function            *)
    mov al, 0       (* Assume FALSE                      *)
    jz  @Done       (* Zero flag set? Yes, then no input *)
    inc al          (* Key was pressed; return TRUE      *)
  @Done:
  END;


FUNCTION ReadKey: char; assembler;
(* Returns ASCII code of key pressed, unless it's special key *)
  ASM
    mov ah, 10h     (* Read character pressed on keyboard  *)
    int 16h         (* Call Keyboard function              *)
  END;


PROCEDURE Writeln (Str: string);
(* This procedure reduces the size of compiled code and can even (in some
   cases) reduce execution time of program. Really! *)
  BEGIN
    system.writeln (Str)
  END;


PROCEDURE PWrite (Str: string);
(* Used for all output to screen except quote display. The current row on the
   screen being displayed is kept track off so before the screen scrolls up,
   the user is prompted to continue.                                     *)
  VAR
    Ch: char;                    (* character pressed by user            *)
  BEGIN
    inc     (CurrRow);           (* increment # of rows displayed so far *)
    Writeln (Str);               (* output string                        *)

    IF (CurrRow = NumRows) THEN  (* time to prompt user?                 *)
      BEGIN
        write ('Press any key to continue...'#13);
        Ch := ReadKey;
        clreol;
        IF (Ch = ESC) THEN       (* halt program if <Esc> key pressed    *)
          BEGIN
            Writeln ('');
            halt (0)
          END
        ELSE IF (Ch = #00) THEN
          Ch := ReadKey;
        CurrRow := 0
      END

  END; (* PWrite *)


PROCEDURE DisplayTitle;
(* Display program's title if it isn't already displayed. *)
  BEGIN
    IF TitleNotShown THEN
      BEGIN
        NumRows := hi (CRT.windmax);
	PWrite (Title);
	TitleNotShown := FALSE
      END
  END;


PROCEDURE DisplayOptions;
(* Display command-line options. *)
  BEGIN
    PWrite ('');
    PWrite ('Syntax: UTTER  [option(s)]  [datafile path]');
    PWrite ('');
    PWrite ('Options: /N ----------- show Number of quotes in data file');
    PWrite ('         /R ----------- display Random quote');
    PWrite ('         /U ----------- Update index file');
    PWrite ('         /W ----------- disable all Warning prompts');
    PWrite ('         /B ----------- use BIOS routines for video output (slower)');
    PWrite ('         /V ----------- use DOS routines for Video output (slowest)');
    PWrite ('         /S ----------- enable video Snow-checking (for CGA monitors)');
    PWrite ('         /K ----------- prompt user to press a Key after quote');
    PWrite ('         /M ----------- prompt user to click Mouse button after quote');
    PWrite ('         /P[x] -------- Pause for approximately x seconds after quote');
    PWrite ('                        (default is 2 seconds. Max. is 60.)');
    PWrite ('         /E[editor] --- invoke Editor (' + Editor + ' is default)');
    PWrite ('         /?, /H ------- for more information')
  END;


PROCEDURE DisplayHelp;
(* Display detailed explanation of user options. *)
  BEGIN
    DisplayTitle;
    DisplayOptions;

    PWrite ('');
    PWrite ('- UTTER displays a quote from an ASCII text data file, called UTTER.DAT');
    PWrite ('  by default, which is assumed to be in the same directory as UTTER.EXE. ');
    PWrite ('');
    PWrite ('- All quotes in the data file must end with a tilde (~); and the length');
    PWrite ('  of each quote is unlimited, as is the size of the data file.');
    PWrite ('');
    PWrite ('- To use a different data file just specify its full path name on the');
    PWrite ('  command-line. If the default data file is in a different directory');
    PWrite ('  than the executable file, then specify the directory name on the');
    PWrite ('  command-line, making sure it ends with a "\" character.');
    PWrite ('');
    PWrite ('- An index file is associated with each different data file, having the');
    PWrite ('  same name as the data file except with an ' + IndexExt + ' extension.');
    PWrite ('');
    PWrite ('- Each index file must be in the same directory as its corresponding data');
    PWrite ('  file. Index files are created automatically if they do not exist in the');
    PWrite ('  same directory as its data file or if they are corrupt in some way.');
    PWrite ('');
    PWrite ('- Options can appear in any order and can be combined.');
    PWrite ('');
    PWrite (' Explanation of options (examples follow)');
    PWrite (' ======================');
    PWrite ('   /N  --------- displays the number of quotes in the current data file.');
    PWrite ('   /R  --------- display a random quote from the current data file.');
    PWrite ('   /U  --------- updates current data file, creating an index file for it.');
    PWrite ('   /W  --------- disables any warning prompts when creating index file.');
    PWrite ('   /B  --------- by default, UTTER outputs directly to video memory. This');
    PWrite ('                 option enables BIOS output routines, which is slower.');
    PWrite ('   /V  --------- use DOS routines for output. This is the slowest but it');
    PWrite ('                 allows any output to be redirected, if desired.');
    PWrite ('   /S  --------- enables snow-checking, which CGA screens may require.');
    PWrite ('   /K  --------- after a quote is displayed, the user is prompted with the');
    PWrite ('                 following: "Press any key to continue...".');
    PWrite ('   /M  --------- after a quote is displayed, the user is prompted with the');
    PWrite ('                 following: "Click mouse button to continue...". If the /K');
    PWrite ('                 option is enabled in addition to this one, the prompt is:');
    PWrite ('                 "Press any key or click mouse button to continue...".');
    PWrite ('   /P[x]  ------ after a quote is displayed, UTTER pauses for the specified');
    PWrite ('                 number of seconds, or for 2 seconds by default. Seconds must');
    PWrite ('                 be >= 0 and <= 60. If x is specified, no other options can');
    PWrite ('                 immediatedly follow x. Ex: "/P3V" -> wrong, "/P3 /V" -> right.');
    PWrite ('   /E[editor]  - executes, by default, the editor EDIT using either the default');
    PWrite ('                 data file UTTER.DAT or the file specified on the command-line.');
    PWrite ('                 To use a different editor, specify the editor''s name');
    PWrite ('                 immediately after the /E option.');
    PWrite ('');
    PWrite (' Examples');
    PWrite (' ========');
    PWrite ('1) UTTER  quotes.txt  /vr  /k');
    PWrite ('   - displays a random quote from the file QUOTES.TXT using DOS output');
    PWrite ('     routines and prompts user to press a key afterwards.');
    PWrite ('');
    PWrite ('2) UTTER  /wu  /eMyEdit');
    PWrite ('   - executes the editor MYEDIT on the default data file UTTER.DAT, then');
    PWrite ('     updates the index file automatically after MYEDIT execution halts.');
    PWrite ('');
    PWrite ('3) UTTER  /sp5  \utils\  /r');
    PWrite ('   - displays a random quote from the default data file UTTER.DAT located');
    PWrite ('     in the \UTILS\ directory using snow-checking, pausing for five seconds');
    PWrite ('     afterwards.');
    PWrite ('');
    PWrite (' Copyright Information');
    PWrite (' =====================');
    PWrite ('- UTTER may be distributed freely in the public domain, although donations');
    PWrite ('  would be most appreciated.');
    PWrite ('');
    PWrite ('- I take no responsibility for any liability, loss, or damage caused or');
    PWrite ('  alleged to be caused directly or indirectly from the use of this program.');
    PWrite ('');
    PWrite ('- UTTER was written in Turbo Pascal 6.0. Source code is available upon request.');
    PWrite ('  Future versions are forthcoming.');
    PWrite ('');
    PWrite ('- If you discover any bugs, or have any ideas or suggestions, please feel');
    PWrite ('  free to contact me:  Jody R. Cairns');
    PWrite ('                       P.O. Box 5991');
    PWrite ('                       St. John''s, NFLD.');
    PWrite ('                       Canada  A1C 5X4      e-mail: jodyc@cs.mun.ca');

    halt (0)
  END;


FUNCTION Num2Str (Number: longint): NumStr;
(* Return Number as a string *)
  BEGIN
    str (Number:1, Param);
    Num2Str := Param
  END;


PROCEDURE CheckError (ErrorCode: word);
(* Display error message and halt program if an error occurs. *)
  VAR
    Error: string;    (* error to be displayed                *)
  BEGIN
    CASE ErrorCode OF
       00: exit;
       02: Error := 'Could not find quote file ' + fexpand (DataPath);
       03: Error := 'Path not found.';
       04: Error := 'Too many open files.';
       05: Error := 'File access denied.';
       06: Error := 'Invalid file handle.';
       08: Error := 'Not enough memory to run ' + EditorPath;
       20: Error := 'Invalid DOS version. Requires DOS 3.0 or greater.';
       21: Error := 'Quote file cannot have ' + IndexExt + ' extension. Rename quote file.';
       22: Error := 'Invalid command-line. No option specified.';
       24: Error := 'Invalid editor name. ' + Editor + ' is not executable.';
       25: Error := 'No editor filename specified.';
       26: Error := 'Could not find editor ' + Editor + Ext;
       27: Error := 'Invalid command-line option: "' + Param[J] + '"';
       28: Error := 'Pause seconds must be >= 0 and <= 60.';
       29: Error := 'Index file not updated or created.';
       30: Error := 'No mouse driver installed on system.';
      100: Error := 'Disk read error. Index file may be corrupt.' +
                    #13#10'       Recommend updating file using /U option.';
      101: Error := 'Disk write error. Disk is probably full.';
      106: Error := 'Invalid numeric format. Index file maybe corrupt.' +
                    #13#10'       Recommend updating file using /U option.';
      150: Error := 'Disk is write-protected.';
      151: Error := 'Unknown disk unit ID.';
      152: Error := 'Drive not ready.';
      154: Error := 'CRC error in data.';
      155: Error := 'Bad drive request structure length.';
      156: Error := 'Disk seek error.';
      157: Error := 'Unknown media type.';
      158: Error := 'Sector not found.';
      160: Error := 'Device write fault.';
      161: Error := 'Device read fault.';
      162: Error := 'Hardware failure.'
      ELSE
	Error := '#' + Num2Str (ErrorCode) + '. Please report this to author.'
    END; (* CASE *)

    DisplayTitle;
    writeln ('ERROR: ' + Error);

    IF (ErrorCode > 20) AND (ErrorCode <= 30) THEN (* command-line error *)
      DisplayOptions;

    halt (ErrorCode)

  END; (* CheckError *)



FUNCTION UpCaseStr (Str: ComStr): ComStr; assembler;
(* Convert Str into uppercase letters and return it.  Code adapted from
   Turbo Pascal 6.0 Programmer's Guide by Borland, pages 305-306.     *)
  ASM
    mov dx, ds              (* Save DS register        *)
    cld                     (* Forward string-ops      *)
    lds si, Str             (* Load string address     *)
    les di, @Result         (* Load result address     *)
    lodsb                   (* Load string length      *)
    stosb                   (* Copy to result          *)
    xor ah, ah              (* String length to cx     *)
    mov cx, ax
    jcxz @3                 (* Skip if empty string    *)
  @1:
    lodsb                   (* Load character          *)
    cmp al, 'a'             (* Skip if not 'a'..'z'    *)
    jb @2
    cmp al, 'z'
    ja @2
    sub al, 20h             (* Convert to uppercase    *)
  @2:
    stosb                   (* Store in result         *)
    loop @1                 (* Loop for all characters *)
  @3:
    mov ds,dx               (* Restore DS register     *)
  END; (* UpCaseStr *)


FUNCTION Continue (Str: string): boolean;
(* Prompt user if they want to continue or not. Return TRUE if they want to;
   otherwise, return FALSE. *)
  VAR
    Ch : char;    (* Character pressed by user *)
  BEGIN
    write (Str, '. '#13#10'Continue (Y/N)? ');
    REPEAT
      Ch := upcase (ReadKey);
      IF (Ch = ESC) THEN
        BEGIN
          Writeln ('');
          CheckError (29)
        END
      ELSE IF (Ch = #00) THEN
        Ch := ReadKey

    UNTIL (Ch = 'Y') OR (Ch = 'N');

    Writeln (Ch);
    Continue := (Ch = 'Y')
  END; (* Continue *)


PROCEDURE RunEditor;
(* Search for editor to be used for editing quote file and execute it *)
  VAR
    Dir    : DirStr;    (* Directory location of editor *)
    Name   : NameStr;   (* Filename of editor           *)
    OrigExt: ExtStr;    (* Extension of editor          *)
    Path   : string;    (* Copy of environment's PATH   *)
  BEGIN
    DisplayTitle;

    Path := Dir + ';' + getenv ('PATH');

    fsplit (Editor, Dir, Name, OrigExt);

    (* see if Name is valid *)
    IF (length (Name) = 0) THEN
      CheckError (25);

    (* Check if user put on extension on editor name *)
    IF (length (OrigExt) = 0) THEN
      Ext := '.COM'

    (* Must be executable file *)
    ELSE IF (OrigExt <> '.COM') AND (OrigExt <> '.EXE') AND (OrigExt <> '.BAT') THEN
      CheckError (24)

    ELSE
      Ext := OrigExt;

    Editor := Dir + Name;

    (* Search for user's editor. Here, it either has a COM extension or
       no extension. *)
    EditorPath := fsearch (Editor + Ext, Path);

    (* Search for editor *)
    IF (length (EditorPath) = 0) THEN  (* didn't find COM editor or users *)

      IF (length (OrigExt) = 0) THEN
        BEGIN
          EditorPath := fsearch (Editor + '.EXE', Path);

          IF (length (EditorPath) = 0) THEN (* didn't find EXE editor *)
            BEGIN
              EditorPath := fsearch (Editor + '.BAT', Path);

              IF (length (EditorPath) = 0) THEN (* didn't find BAT editor *)
                BEGIN
                  Ext := OrigExt;
                  CheckError (26)
                END
            END
        END
      ELSE
        CheckError (26);

    EditorPath := fexpand (EditorPath);

    (* execute editor *)
    swapvectors;
    exec (getenv ('COMSPEC'), '/C ' + Editor + ' ' + DataPath);
    swapvectors;
    CheckError (DOSError)

  END; (* RunEditor *)


PROCEDURE CheckIOError (Error: word);
(* This procedure is only used by UpdateIndexFile to check for
   Input/Output errors when reading and writing is done to the data file
   and index file. *)
  BEGIN
    IF (Error <> 0) THEN
      BEGIN
        writeln ('');
        CheckError (Error)
      END
  END; (* CheckIOError *)


PROCEDURE UpdateIndexFile (Code: word);
(* Update the index file for the particular quote file in use.
   The quote file (DataFile) is scanned for the end of quote marker (EndOfQuote),
   and each position of EndOfQuote is written to IndexFile. The complete
   format of IndexFile is as follows (records are longint):

     1st record (OldDataTime):
     --- last recorded date and time of DataFile. This is compared with
         the current date and time of DataFile to see if DataFile has
         been altered in any way. If it has, then IndexFile is updated.

     2nd record (StartPos):
     --- seek position into DataFile where next quote that is to be
         displayed is located. For example: Let StartPos := 274
         Then before a quote is displayed, we "seek (DataFile, StartPos)",
         which moves the file pointer for DataFile to position 274 in
         DataFile, which is where the quote to be displayed is located.

     3rd and remaining records:
     --- these records are similar to the 2nd record above except that they
         are the sequential positions the each quote in DataFile. Thus, the
         3rd record is always equal to zero, since that marks the beginning
         of the first quote in DataFile. The 4th record indicates the next
         location of the 2nd quote in DataFile, and so on.
         These records are ONLY used by the UTTER /R option, or the Random
         option.  If that option is chosen, UTTER "seek"s to a random
         record in IndexFile from the 3rd to last record. Then, StartPos
         is read from IndexFile to indicate where the quote in DataFile is
         located.
                                                                           *)

  CONST
    NumberOfQuotes: longint = 0;

  BEGIN
    IF DoPrompts AND (NOT Continue ('About to create a new index file for ' + fexpand (DataPath))) THEN
      CheckError (Code);

    IF (UpCaseStr (Ext) = IndexExt) THEN  (* Datafile can't have .IDX extension *)
      CheckError (21);

    UpDateFile := TRUE;

    writeln ('Creating ' +  fexpand (Dir + Name) +  IndexExt);

    rewrite    (IndexFile);
    CheckError (IOResult);

    write (#13'Number of quotes: ', NumberOfQuotes:1);

    write (IndexFile, DataTime, NumberOfQuotes, NumberOfQuotes);
    CheckIOError (IOResult);

    REPEAT
      blockread (DataFile, Buffer, sizeof (Buffer), RecsRead);
      CheckIOError (IOResult);

      FOR I := 1 TO RecsRead DO
        IF (Buffer[I] = EndOfQuote) THEN
          BEGIN
            inc (NumberOfQuotes);
            write (#13'Number of quotes: ', NumberOfQuotes:1);

            StartPos := (filepos (DataFile) - (RecsRead - I));
            write (IndexFile, StartPos);
            CheckIOError (IOResult);
          END
    UNTIL (RecsRead = 0);

    writeln ('');

    OldDataTime := DataTime;
    StartPos    := 0;
    seek (IndexFile, 0)

  END; (* UpdateIndexFile *)


FUNCTION ButtonPressed: boolean; assembler;
(* The mouse button status is returned in register bx; bl = 0 means no buttons
   are being pressed, 1 means left button, 2 = right button, and 3 means
   both buttons. We don't care which buttons here.                          *)
  ASM
    push bx        (* Save registers changed *)
    push cx
    push dx
    mov  ax, 03h   (* Get mouse status       *)
    int  33h       (* Call mouse function    *)
    mov  al, bl    (* Return result          *)
    pop  dx        (* Restore registers      *)
    pop  cx
    pop  bx
  END;


FUNCTION NoMouse: boolean; assembler;
(* Returns TRUE if there is no mouse driver installed; otherwise, FALSE.
   To check if a mouse alone is installed takes too long, even though
   it's simply one function call.
   First, we see if there is a vector address for Int 33h, the mouse
   interrupt. Then, if there is, we have to make sure that the code
   there isn't just the instruction IRET (CFh in op-code).  Adapted from
   Allen L. Wyatt, Sr.'s Advanced Assembly Language, pages 373-374.         *)
  ASM
    push bx                       (* Save registers changed         *)
    push es
    mov ah, 35h                   (* Get interrupt vector           *)
    mov al, 33h                   (* Mouse interrupt                *)
    int 21h                       (* Call function                  *)
    or  ax, bx                    (* Was an address returned?       *)
    jz  @No_Mouse                 (* No, so there's no mouse        *)
    cmp byte ptr es:[bx], 0CFh    (* Is IRET at address?            *)
    je  @No_Mouse                 (* Yes, so no driver installed    *)
    xor al, al                    (* Return FALSE - mouse installed *)
    jmp @Done                     (* We're done                     *)
  @No_Mouse:                      (* There's no mouse               *)
    mov al, 1                     (* Return TRUE - no mouse         *)
  @Done:
    pop es                        (* Restore registers              *)
    pop bx
  END; (* NoMouse *)


(*************************************************************************)
(*************************************************************************)


BEGIN (* main program *)

  (* See if valid DOS version *)
  IF (lo (dosversion) < 3) THEN
    CheckError (20);

  (* Turn off snow checking   *)
  CheckSnow := FALSE;

  (* Get directory where UTTER is residing *)
  fsplit (fexpand (paramstr(0)), Dir, Name, Ext);

  Ext := '.DAT';    (* Default quote file extension *)

  (* Process any command-line parameters *)
  FOR I := 1 TO paramcount DO
    BEGIN

      Param := paramstr (I);
      IF (Param[1] = '/') THEN          (* Options are preceded by '/' *)
        BEGIN
          IF (length (Param) = 1) THEN   (* No option specified *)
            CheckError (22);

          FOR J := 2 TO length (Param) DO
            CASE upcase (Param[J]) OF
  	      'N': ShowNumber  := TRUE;
	      'R': GetRandom   := TRUE;
              'U': UpdateFile  := TRUE;
              'W': DoPrompts   := FALSE;
              'S': CheckSnow   := TRUE;
              'B': DirectVideo := FALSE;
              'V': BEGIN (* Use standard DOS output *)
                     assign  (output, '');
                     rewrite (output)
                   END;
	      'E': BEGIN (* Invoke editor *)
                     InvokeEditor := TRUE;
                     IF (J <> length (Param)) THEN  (* get specified editor *)
                       BEGIN
                         Editor := UpCaseStr (copy (Param, J+1, length (Param)));
                         goto EndLoop1
                       END
                   END;
              'P': BEGIN (* Pause after quote is displayed *)
                     Pause  := TRUE;
		     TmpStr := copy (Param, J+1, length (Param));
                     IF (length (TmpStr) <> 0) THEN
                       BEGIN
                         val (TmpStr, LongIntVar, Code);
                         IF (Code <> 0) OR (LongIntVar > MaxSeconds) OR
                            (LongIntVar < 0) THEN
                           CheckError (28);
                         PauseMilSecs := LongIntVar * 1000;
                         goto EndLoop1
                       END
                   END;
              'K': BEGIN (* Prompt user after quote is displayed *)
                     KeyPause := TRUE;
                     IF MousePause THEN
                       Prompt := #13#10'Press any key or click mouse button to continue...'
                   END;
              'M': BEGIN (* Prompt user after quote is displayed *)
                     IF NoMouse THEN
                       CheckError (30);
                     MousePause := TRUE;
                     IF KeyPause THEN
                       Prompt := #13#10'Press any key or click mouse button to continue...'
                     ELSE
                       Prompt := #13#10'Click mouse button to continue...'
                   END;
              '?',
              'H': DisplayHelp
	      ELSE
	        CheckError (27)      (* Invalid command-line option *)
	    END; (* CASE *)
          EndLoop1:
        END (* IF *)

      ELSE (* different data file or directory to be used *)
        BEGIN
          IF (Param = '?') THEN
            DisplayHelp;
          IF (Param[length (Param)] = '\') THEN  (* different directory *)
            Dir := Param
          ELSE
            fsplit (Param, Dir, Name, Ext)       (* different data file *)
        END; (* ELSE *)

    END; (* FOR I *)

  DataPath := Dir + Name + Ext;  (* Quote data file's full path name  *)

  FileMode := 0;                 (* Set file mode to read access only *)

  (* Open data file and see if it is valid *)
  assign (DataFile, DataPath);
  reset  (DataFile, 1);
  CheckError (IOResult);

  IF InvokeEditor THEN  (* Run editor on data file if requested now *)
    RunEditor;

  (* Get current data file's time *)
  getftime (DataFile, DataTime);

  FileMode := 2;       (* Set file mode to read and write access *)

  (* Open index file for data file and see if it is valid *)
  assign (IndexFile, Dir + Name + IndexExt);
  reset  (IndexFile);
  Code := IOResult;

  (* Check if any errors occured when index file was opened or if user  *)
  (* requested to automatically update index file                       *)

  IF (Code <> 0) OR UpdateFile THEN
    BEGIN

      (* Code = 2 means 'file not found'. Anything else we can't deal with *)
      IF (Code <> 2) AND (NOT UpdateFile) THEN
        CheckError (Code);

      DisplayTitle;

      IF UpdateFile THEN   (* Code = 29 -> index file not created *)
        Code := 29;

      (* Create new index file *)
      UpdateIndexFile (Code)

    END; (* IF *)

  (* Read old data file's time and seek position of quote to be displayed *)
  read (IndexFile, OldDataTime, StartPos);
  Code := IOResult;
  IF (Code <> 0) THEN
    BEGIN
      DisplayTitle;

      (* Code = 106 means 'invalid numereic format' and                    *)
      (* Code = 100 means 'disk read error' which both probably indicates  *)
      (* that the index file is corrupt, which we can deal with.   *)

      IF (Code <> 106) AND (Code <> 100) THEN
        CheckError (Code);

      writeln ('Something is amiss...');

      UpdateIndexFile (Code)

    END; (* IF *)

  (* Compare old and new times, and update index file if different *)
  IF (OldDataTime <> DataTime) THEN
    BEGIN
      DisplayTitle;
      writeln ('Something is amiss...');

      UpdateIndexFile (29)

    END; (* IF *)

  (* Display number of quotes in data file *)
  IF ShowNumber AND (NOT UpDateFile) THEN
    BEGIN
      DisplayTitle;
      writeln ('Number of quotes in ' + fexpand (DataPath) + ': ' + Num2Str (filesize (IndexFile) - 3))
    END;

  (* Get random quote. *)
  IF GetRandom THEN
    BEGIN
      randomize;

      StartPos := (filesize (IndexFile) - 1);

      (* StartPos in this case must be <= a word-range because of the  *)
      (* function 'random', which accepts nothing larger than a word.  *)

      IF (StartPos > MaxWord) THEN
        StartPos := MaxWord
      ELSE
        dec (StartPos);

      IF (StartPos <= 0) THEN
        UpDateIndexFile (29);

      seek (IndexFile, (random (StartPos) + 2));
      CheckError (IOResult);
      read (IndexFile, StartPos);
      CheckError (IOResult)
    END; (* GetRandom *)


  (* Seek into data file for the position of the quote to be displayed  *)
  seek (DataFile, StartPos);

  (* Display quote *)
  REPEAT
    blockread  (DataFile, Buffer, sizeof (Buffer), RecsRead);
    CheckError (IOResult);

    FOR I := 1 TO RecsRead DO
      IF (Buffer[I] <> EndOfQuote) THEN
        write (Buffer[I])
      ELSE
        goto EndLoop2

  UNTIL (RecsRead = 0);

  EndLoop2:

  system.writeln;

  (* Prepare to write next quote to be displayed position *)
  seek (IndexFile, 1);

  IF (RecsRead = 0) THEN (* End-of-file, so next quote is first quote in file *)
    StartPos := 0

  ELSE                   (* Save next quote position *)
    StartPos := (filepos (DataFile) - (RecsRead - I));

  (* Write next quote position in index file *)
  write      (IndexFile, StartPos);
  CheckError (IOResult);

  (* Close files *)
  close      (IndexFile);
  CheckError (IOResult);
  close      (DataFile);
  CheckError (IOResult);

  IF Pause THEN
    delay (PauseMilSecs)

  ELSE IF KeyPause OR MousePause THEN
    BEGIN
      writeln (Prompt);                    (* Display prompt *)
      IF KeyPause AND MousePause THEN
        BEGIN
          REPEAT
          UNTIL KeyPressed OR ButtonPressed;
        END
      ELSE IF KeyPause THEN
        REPEAT
        UNTIL KeyPressed
      ELSE
        REPEAT
        UNTIL ButtonPressed;

      WHILE KeyPressed DO    (* Eat up any characters left over in the *)
        Ch := ReadKey        (* keyboard buffer.                       *)
    END

END. (* main program *)